/***
* BSFFT1D is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
* 
* $LastChangedDate: 2016-05-20 13:54:36 +0200 (Fr, 20. Mai 2016) $
* $Author: melzer $
* $Revision: 6398 $
*
*/





#include "sparseFFT1D.h"

sparseFFT1D::sparseFFT1D(vec *MX_input,vec *MOmega_input,unsigned int p_input, unsigned int N_input,string opt){

	if (opt.compare("showCopyInfo")==0){
		showCopyright();
	}


	MX=MX_input;
	MOmega=MOmega_input;
	p=p_input;
	N=N_input;

	unsigned int L=log2(N);

	//tree TX
	unsigned int M1=MX->n_rows;
	vector<unsigned int> * ind1 =new vector<unsigned int>;
	for(unsigned int i=0;i<M1;i++){
		ind1->push_back(i);
	}

	tree1D TX(MX,ind1,L,0,0);

	delete ind1;

	 vector <unsigned int> * index_out1= new vector <unsigned int>;
			TX.getIndicesOfTree(index_out1);
			/* cout << "Alle Indizes des Baumes= " << endl;
			cout << " i " << "M[i]" << endl;
			for (unsigned int k=0; k< index_out1->size();k++){
				cout << " " << index_out1->at(k) << " " << MX->at(index_out1->at(k)) << endl;
			}*/


	//tree TOmega
	unsigned int M2=MOmega->n_rows;

	vector<unsigned int> * ind2 =new vector<unsigned int>;
	for(unsigned int i=0;i<M2;i++){
		ind2->push_back(i);
	}


	tree1D TOmega(MOmega,ind2,L,0,0);
	delete ind2;

	//grids in [-1/2,1/2]

	alpha=zeros<vec>(p);
	beta=zeros<vec>(p);

	for(unsigned int i=0;i<p;i++){
			alpha.at(i)=(0.5*cos((2.*i+1)/(2.*p)*M_PI) );
			beta.at(i)=(0.5- (i*1.)/((p-1)*1.));
	}//for


	//box parameters
	for(unsigned int l=1; l <=L;l++){
		vector<unsigned int> *tmpALevelPara = new vector<unsigned int>;
		TX.getLevelParaA(tmpALevelPara,l);
		//cout << "A " << tmpALevelPara << " " << tmpALevelPara->size() <<  endl;
		ABoxes.push_back(tmpALevelPara);
		vector<structLevelParaB>* tmpBLevelPara = new vector<structLevelParaB>;
		TOmega.getLevelParaB(tmpBLevelPara,L-l);
		//cout << "B " << tmpBLevelPara << " " << tmpBLevelPara->at(1).m1B <<  endl;
		BBoxes.push_back(tmpBLevelPara);
		numBoxesLevel.push_back(TOmega.countSonsInLevel(L-l));
	}//for

		//cout << "Size ABoxes=" <<  sizeof(ABoxes) << endl;
		//cout << "Size BBoxes=" <<  sizeof(BBoxes) << endl;


	//box parameters in level L
	vector<structLevelBox> *tmpAL = new vector<structLevelBox>;
	TX.getBoxes(tmpAL,L);
	lastLevelA=tmpAL;

	/*
	for (int k=0; k< tmpAL->size();k++){
		cout << "m= " <<tmpAL->at(k).mBox << endl ;
		for (int j=0; j < tmpAL->at(k).pind->size();j++){
		cout << "    Index= "<< tmpAL->at(k).pind->at(j)  << " Wert " << MX->at(tmpAL->at(k).pind->at(j))<< endl;
	}
	}*/

	vector<structLevelBox> *tmpBL = new vector<structLevelBox>;
	TOmega.getBoxes(tmpBL,L);
	lastLevelB=tmpBL;


	//Init Lagrange matrices
	 Lleft=zeros<matcomplex>(p,p);
	 Lright=zeros<matcomplex>(p,p);
	 veccomplex * tmp0= new veccomplex;
	 veccomplex * tmp1=new veccomplex;

		for(unsigned int k=0;k<p;k++){

			lagpol_alpha(tmp0,0,k);
			lagpol_alpha(tmp1,1,k);

			Lleft.col(k)=*tmp0;
			Lright.col(k)=*tmp1;
		}
		delete tmp0;
		delete tmp1;

}//constructor

sparseFFT1D::~sparseFFT1D(){

	delete lastLevelA; //cellarray: last level of boxes A from T_X with rows:  {m,indices,(number of sons)}
	delete lastLevelB; //cellarray: last level of boxes B from T_Omega with rows: {m,indices,number of sons}

	for(unsigned int i=0; i<ABoxes.size();i++){
			delete ABoxes.at(i);
		}
	for(unsigned int i=0; i<BBoxes.size();i++){
				delete BBoxes.at(i);
			}

}//destructor

/**
 * vector z: evaluation points
 * vector z_nodes: interpolation nodes
 * parameter k for the k-th Lagrange polynome
 */

void sparseFFT1D::lagpol_z(veccomplex *pls,veccomplex *z,veccomplex *z_node,unsigned int k)
{

	//complex<double> ls(0,1);

	veccomplex ls = ones< veccomplex >(p);


	unsigned int j;
	for(j=0; j<k;j++){
		complex<double> z_j = z_node->at(j);
		ls=ls % (( (*z)-z_j)/(z_node->at(k)-z_j));
	}
	for(j=k+1; j<p;j++){
		complex<double> z_j = z_node->at(j);
		ls=ls%(((*z)-z_j)/(z_node->at(k)-z_j));
	}

	*pls=ls;

}

/**
 *  vector *pls =[l_s(x_1),...,l_s(x_d)]
 *  sonType=0 for the left and sonType=1 for the right son
 *  parameter k for the s-th lagrange polynome
*/
void sparseFFT1D::lagpol_alpha(veccomplex *pls,unsigned int sonType, unsigned int k){
	veccomplex ls = ones< veccomplex >(p);
	veccomplex z(p);
	veccomplex z_node(p);

	unsigned int j;
	for (j=0;j<p;j++){
		z(j)=exp(pow(-1,sonType)*M_PI*I/(2*(double) (p-1)))*exp(-M_PI*I*alpha.at(j)/ (double) (p-1));
		z_node(j)=exp(-2*M_PI*I*alpha.at(j)/(double) (p-1));
	}
	dcomplex z_k=z_node(k);


		for(j=0; j<k;j++){
			complex<double> z_j = z_node(j);
			ls=ls % (( z-z_j)/(z_k-z_j));
		}

		for(j=k+1; j<p;j++){
		   complex<double> z_j = z_node(j);
		   	ls=ls % (( z-z_j)/(z_k-z_j));
	}
	*pls=ls;
}//lagpol_alpha


void sparseFFT1D::showCopyright(){
	printf("COPYRIGHT INFOS...\n");
		printf("	BSFFT1D (C) 2016 Ines Melzer\n");
		printf("	\n");
		printf("	BSFFT1D is free software: you can redistribute it and/or modify\n");
		printf("	it under the terms of the GNU General Public License as published by\n");
		printf("	the Free Software Foundation, either version 3 of the License, or\n");
		printf("	(at your option) any later version.\n");
		printf("\n");
		printf("	This program is distributed in the hope that it will be useful,\n");
		printf("	but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		printf("	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		printf("	GNU General Public License for more details.\n");
		printf("\n");
		printf("	You should have received a copy of the GNU General Public License\n");
		printf("	along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
		printf("\n");
}
