/***
* BSFFT1D is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
* 
* $LastChangedDate: 2016-05-20 13:54:36 +0200 (Fr, 20. Mai 2016) $
* $Author: melzer $
* $Revision: 6398 $
*
*/





#include "sparseFFT1D_mtimes.h"


sparseFFT1D_mtimes::sparseFFT1D_mtimes(sparseFFT1D * plan_input, veccomplex *f_input){
	plan=plan_input;
	fhat=f_input;
	//veccomplex tmpf = zeros<veccomplex>(plan->MX->n_rows);
	//f=&tmpf;
	f=new veccomplex(plan->MX->n_rows);
}

void sparseFFT1D_mtimes::setf(veccomplex *f_input){

	(*f)=(*f_input);
}


sparseFFT1D_mtimes::~sparseFFT1D_mtimes(){

}

void sparseFFT1D_mtimes::mtimes(){
	unsigned int L=log2(plan->N);
	double cA=(double) plan->N/2;
	double wA=(double) plan->N;

	vec xA=cA+plan->alpha*wA; //grid in A
	unsigned int mA=0;
	vector<structLevelBox> *BBoxes=plan->lastLevelB;

	//computing function values for l=0, A=X_00

	//U_cell=zeros(plan.p+2,size(B_Boxes,1));
	vector <cellLevelBox> * U_prev = new vector<cellLevelBox>;
	for(unsigned j=0; j < BBoxes->size(); j++){
		//parameter of the current box B
		structLevelBox tmp=BBoxes->at(j);
		unsigned int mB=tmp.mBox;
		vector < unsigned int> * Bindex=tmp.pind;

		//computing the function values of the exact functions u^B
		veccomplex * u_AB=new veccomplex;
		getFirstFuncVal1D(u_AB,wA,cA,Bindex);

		cellLevelBox tmpU;
		tmpU.mA=mA;
		tmpU.mB=mB;
		tmpU.u_AB=u_AB;
		//cout << mA << " " << mB << " " << *u_AB << endl;
		U_prev->push_back(tmpU);

	}//for


	//computing new function values from recursive from the previous level

	vector <cellLevelBox> * U_new = new vector<cellLevelBox>;
	vector <cellLevelBox> * tmpIN = U_prev;

	for(unsigned int l=1;l<=L;l++){
					U_new->resize(plan->ABoxes.at(l-1)->size() * plan->BBoxes.at(l-1)->size()); //set length of vector
					//cout << U_new->size() << endl;
		            cout << "	level = " << l << endl;
	                unsigned int numSonsLevel_l=plan->numBoxesLevel.at(l-1);
	                getFuncVal1DLevel(U_new,U_prev,plan->ABoxes.at(l-1) ,plan->BBoxes.at(l-1),l,numSonsLevel_l);

	                //change pointers
	                tmpIN=U_prev;
	                U_prev=U_new;
	                U_new=tmpIN;

	}//for l

	cout << "	Evaluation in X nodes" << endl;

	//last step: evaluation in the nodes MX using barycentric interpolation formula

	double cB=(double) plan->N/2;
	double wB=(double) plan->N;
	wA=1;
	veccomplex u=zeros<veccomplex>(plan->MX->n_rows);

	vector<structLevelBox> * ABoxes=plan->lastLevelA; //A-Boxes of level L
	dcomplex c2piiN=2*M_PI*I/((double) plan->N);
	dcomplex c2piiNpwB=2*M_PI*I*wB/((double) (plan->p-1)*plan->N);
	veccomplex denom(plan->p); // denominator of lagrange polynomial
	veccomplex zalpha(plan->p);
	zalpha=exp(-2*M_PI*I*plan->alpha/( (double) plan->p-1));

	for(unsigned int s=0;s<plan->p;s++){
	   denom(s)=1;
	   for(unsigned int j=0;j<s;j++){
		   denom(s)=denom(s)*(zalpha(s)-zalpha(j));
	   }
	   for(unsigned int j=s+1;j<plan->p;j++){
	   		   denom(s)=denom(s)*(zalpha(s)-zalpha(j));
	   	   }
	}//for s

	veccomplex unitp=ones<veccomplex>(plan->p-1);

	for(unsigned int r=0; r < U_prev->size(); r++){

		unsigned int mA=U_prev->at(r).mA;
	    veccomplex * u_AB=U_prev->at(r).u_AB;
	    vector<unsigned int> * Aindex=ABoxes->at(r).pind;

	    cA=wA*(mA+0.5);
	    vec xA=cA+plan->alpha*wA;

	    veccomplex d(plan->p);
	    d=exp(-c2piiN*xA*(cB+wB/2));

	    int numA=Aindex->size();
	    vec x(numA);
	    for(int j=0;j<numA;j++){
	    	x(j)=plan->MX->at(Aindex->at(j));
	    }

	    veccomplex e(numA);
	    veccomplex z(numA);
	    e=exp(c2piiN*x*(cB+wB/2));
	    z=exp(-c2piiNpwB*(x-cA));

	    matcomplex tmpz=unitp*(z.st());
	    veccomplex unitA=ones<veccomplex>(numA);
	    veccomplex tmpzalpha(plan->p-1);
	    matcomplex lag=zeros<matcomplex>(numA,plan->p);

	    for (unsigned int s=0;s<plan->p;s++){

	    	for(unsigned int t=0;t<s;t++){
	    			   tmpzalpha(t)=zalpha(t);
	    	}
	    	for(unsigned int t=s;t<plan->p-1;t++){
	    			   tmpzalpha(t)=zalpha(t+1);
	    	}
	    	matcomplex Z = tmpzalpha*unitA.st();
	    	veccomplex numerator = vectorise(tmpz) - vectorise(Z);
	    	veccomplex numer= (prod( reshape(numerator,plan->p-1,numA),0 )).st(); //numerator of lagrange polynomial, row-vector
	    	lag.col(s)=numer/denom(s);

	    }//for s

	    matcomplex tmpU = (e % (lag*(d % (*u_AB)) ));
	    for(int j=0;j< numA; j++){
	    	u(Aindex->at(j)) = tmpU(j);
	    } //for j
	} // for r, lastLevelA


	setf(&u);

	delete U_new;
	delete U_prev;



}//mtimes


/**
 * Output: exact function evaluated in the chebyshev nodes xA
 */

void sparseFFT1D_mtimes::getFirstFuncVal1D(veccomplex *u ,double w, double c, vector<unsigned int> *Bindex){
	veccomplex tmpu= zeros<veccomplex>(plan->p);
	vec xA=c+(plan->alpha) * w; //Chebyshev grid in A
	for(unsigned int t=0;t< plan->p;t++){
		for(unsigned int j=0;j< Bindex->size();j++){
			double xA_t=xA.at(t);
			double xi_j=plan->MOmega->at(Bindex->at(j));

			//cout << Bindex->at(j) << endl;

			tmpu(t)=tmpu(t)+exp(2*M_PI*I*xA_t*xi_j/(double) plan->N)*( (*fhat)(Bindex->at(j)) );
		}
	}
	*u=tmpu;
}//getFirstFuncVal1D

void sparseFFT1D_mtimes::getFuncVal1DLevel(vector <cellLevelBox> * U_new, vector <cellLevelBox> * U_prev, vector< unsigned int> * AParam, vector< structLevelParaB > * BParam, unsigned int l, unsigned int numSonsLevel_l){

		dcomplex c2piiN=2*M_PI*I/(double) plan->N;

		double mA=(double) AParam->at(0);
		double mP=floor((double) mA/2);

		unsigned int parent=0;//counter

		double wA=(double) plan->N/pow(2,l);
		double wP=(double) plan->N/pow(2,l-1); //P is parent box of A
	    double wS=plan->N/pow(2,log2(plan->N)-l+1); //S is a son box of B

		for(unsigned int j=0; j < AParam->size();j++){
			//Loop over boxes A
			unsigned int mA=AParam->at(j);
			double cA=wA*(mA+0.5);

			vec xA=cA+(plan->alpha) * wA; //Chebyshev grid in A

			//test, if the parent box changes
	        if (  (j!=0) && ( mP !=floor((double) mA/2)   )){
				++parent;
				mP=floor((double) mA/2);
				//cout << "different parent box" << endl;
	        }
	        //else{ cout << "same parentbox" << endl;}

	        matcomplex Lag;
	        //test, if A is the left or the right son
	        if ( ((int) mA % 2)==0){
	           Lag=plan->Lleft;}
	        else{
	            Lag=plan->Lright;}
	        unsigned int col=parent*numSonsLevel_l;

			double cP=wP*(mP+0.5); //center point of the parent box P
			vec xP=cP+(plan->alpha) * wP;//Chebyshev grid in parent box P

			//compute u^{AB} from u^{PS}
	 		for(unsigned int k=0; k < BParam->size(); k++){
	 			//Loop over B
	 			structLevelParaB tmpB = BParam->at(k);
				unsigned int mB=tmpB.m1B;
				unsigned int numSonsB=tmpB.c; //number of sons of B

				veccomplex * tmpu_AB = new veccomplex(plan->p);
				(*tmpu_AB)=zeros<veccomplex>(plan->p);

				for (unsigned int r=0; r < numSonsB; r++){
					//Loop over sons of B
					veccomplex * u_PS= U_prev->at(col+r).u_AB;

					//box parameters of the son
	                unsigned int mS=U_prev->at(col+r).mB;
					double cS=wS*(mS+0.5);

					//column vectors of diagonal matrices R^{AS} and S^{PS}
					veccomplex r_AS(plan->p);
					veccomplex s_PS(plan->p);
					r_AS=exp(c2piiN* xA * (cS+wS/2));
					s_PS=exp(-c2piiN* xP* (cS+wS/2));
					veccomplex tmpu=r_AS % (Lag*(s_PS % (*u_PS)) );
					(*tmpu_AB)+=tmpu;

				}//for r, numSonsB

				col+=numSonsB;

				cellLevelBox tmpBox;
				tmpBox.mA=mA;
				tmpBox.mB=mB;
				tmpBox.u_AB=tmpu_AB;
				U_new->at(j*BParam->size() +k)=tmpBox;

	 		}//for B


		}//for A

}//getFuncVal1DLevel
