/***
* BSFFT1D is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
* 
* $LastChangedDate: 2016-05-20 13:54:36 +0200 (Fr, 20. Mai 2016) $
* $Author: melzer $
* $Revision: 6398 $
*
*/






#include "tree1D.h"

/**
 * Decomposed given input points regarding to the subintervals of m_input
 *
 * @param index_output vector of pointers to the left and right son of X_l,m
 * i.e., decompose interval m in two similar subintervals and assign the input points
 * in index_input to the subintervals
 * @param tmpM input points
 * @param index_input indices of input points inside the interval m in current level l
 * @param l current level
 * @param m current interval in given level
 * @param L non-harmonic bandwith N=2^L
 */
void tree1D::decomposeIndex1D(vector<vector<unsigned int>*> *index_output, vec *tmpM, vector<unsigned int> *index_input,unsigned int l_input, unsigned int m_input, unsigned int L_input){
	double N_input=(double) pow(2,L_input);
	for(unsigned int j=0;j < index_input->size();j++){

		//cout << j << endl;
		//cout << "index_input[" << j <<"] =" << index_input->at(j) << endl;
		double tmp=tmpM->at(index_input->at(j));
	//	cout << "M[j]=" << tmp << endl;

		int k;
		//cout << "const = " << (double)N_input/pow(2,l_input)*((double)m_input+ 0.5) << endl;

		if (tmp < N_input/pow(2,l_input)*(m_input+0.5) ){k=0;} //left son
		else {k=1;} //right son

		//cout << "k=" << k << endl;
		if (index_output->at(k)==NULL){
			vector<unsigned int> *tmpidx = new vector<unsigned int>;
			index_output->at(k)=tmpidx;
		}
		vector<unsigned int> *tmpidx = index_output->at(k);
		//cout << "tmpidx= " << tmpidx << endl;

		tmpidx->push_back(index_input->at(j));

	//	cout << index_input->at(j)<< endl;


	}//for
	//cout << "DCI son 0=" << index_output->at(0) << endl;
	//cout << "DCI son 1=" << index_output->at(1) << endl;
	 //cout << "---" << endl;
}

tree1D::tree1D(){
	l=0;
	m=0;
	L=0;
	pind=0;
	numOfSons=0;
	sons.push_back(NULL);
}
tree1D::tree1D(vec *M, vector<unsigned int> *pind_input, unsigned int L_input, unsigned int l_input, unsigned int m_input){
	l=l_input;
	m=m_input;
	L=L_input;
	pind=NULL;

	if (l==L_input){
		//Level L
		//save indices
		pind=pind_input;
		//cout << "l="  << l_input << "  m=" << m_input << endl;
		for (unsigned int k=0; k < pind->size();k++){
			//cout << pind->at(k) << endl;
			//cout << pind << " " << pind->size() << endl;
		}
		numOfSons=0; //last Level -> no more sons
	}//if
	else{
		//cout << pind << endl;
		//decompose nodes and create sons
		vector<vector<unsigned int>*> *sons_index = new vector<vector<unsigned int> *>(2);
		decomposeIndex1D(sons_index,M,pind_input,l,m,L_input);

		//cout << "Son 0 = " << sons_index->at(0) << endl;
		//cout << "Son 1 = " << sons_index->at(1) << endl;
		numOfSons=0;
		int k;
		for(k=0;k<2;k++){
			if(sons_index->at(k) != NULL){
				tree1D * tmpson = new tree1D(M,sons_index->at(k),L_input,l+1,m*2+k);
				//	cout << tmpson << endl;
				sons.push_back(tmpson);
				//	cout << "sons[k]=" <<  sons.at(k) << endl;
				numOfSons+=1;
			}//if
			//else sons.push_back(NULL);
			//cout << sons.size() << endl;
		}//for
		}//else
}


tree1D::~tree1D(){
		//cout << "Delete Tree" << endl;
		//cout << "l= " << l  << " m= " << m << " numOfSons= " << numOfSons << endl;

		for(unsigned int i=0;i<numOfSons;i++){
			delete sons.at(i);
		}
}



/**
 * output: number of all sons in level l == number of boxes in level l+1
 * @param l_input
 * @return
 */
unsigned int tree1D::countSonsInLevel(unsigned int l_input)
{
	unsigned int out=0;
	if(l==l_input-1){
		unsigned int i;
		for(i=0; i<numOfSons;i++){
			out+=sons.at(i)->numOfSons;
			//cout << sons.at(i)->numOfSons << " " <<   sons.at(i)->sons.size() << endl;
			//cout << sons.at(0) << "  " << sons.at(1) << endl;
		}
	}else if(l==l_input){
		out=numOfSons;
	}else if((l+1)<l_input){
		unsigned int i;
		for(i=0; i<numOfSons;i++){
			out+=sons.at(i)->countSonsInLevel(l_input);
		}
	}
	return out;
}

/**
 * @param vector of positions in level l (list of [m])
 * @param l_input requested level
*/

void tree1D::getLevelParaA(vector<unsigned int> *levelPara, unsigned int l_input){
	if (l==l_input){
		levelPara->push_back(m);
	}//if
	else {
		for(unsigned int k=0; k < numOfSons; k++){
			sons.at(k)->getLevelParaA(levelPara,l_input);
		}//for
	}//end
}//getLevelParaA


/**
 * @param vector of structLevelParaB ([m,c]) positions and their number of sons in level l
 * @param l_input requested level
 */

void tree1D::getLevelParaB(vector<structLevelParaB> *levelParaB,unsigned int l_input){
	if (l==l_input){
		structLevelParaB * tmp= new structLevelParaB;
		tmp->m1B=m;
		tmp->c=numOfSons;
		levelParaB->push_back(*tmp);
		delete tmp;
	}//if
	else{
		for(unsigned int k=0; k < numOfSons; k++){
			sons.at(k)->getLevelParaB(levelParaB,l_input);
		}//for
	}//else

}//getLevelParaB


/*
 * Output: all boxes of level l as cellaray with rows {m,indices,number of sons}
 */
void tree1D::getBoxes(vector<structLevelBox> *box, unsigned int l_input){

  if(l==l_input){

	  if(pind!=NULL){
    	  structLevelBox tmp;
    	  tmp.numOfSons=numOfSons;
    	  tmp.mBox=m;
    	  tmp.pind=pind;
    	  box->push_back(tmp);
    	  //delete tmp;
      }
      else{
          vector<unsigned int> * index_out = new vector <unsigned int>;
          getIndicesOfTree(index_out);
          structLevelBox tmp;
          tmp.numOfSons=numOfSons;
          tmp.mBox=m;
          tmp.pind=pind;
          box->push_back(tmp);
          //delete tmp;
      }
  }
  else{
      for(unsigned int k=0; k < numOfSons;k++){
          sons.at(k)->getBoxes(box,l_input);
  	  }
  }
}//getBoxes

/**
 * Test functions
 */


/**
 * output: all indices which lie in the boxes of the tree X
 *
 * @param indexOut
 */
void tree1D::getIndicesOfTree(vector <unsigned int> * index_out)
{
	//cout << "get Indices" << endl;

	if(pind !=NULL){
		for(unsigned int k=0; k < pind->size();k++){
					//	cout << "Element" << pind->at(k) << endl;
						index_out->push_back(pind->at(k)); //concatenate
					}//for
	}else{
		//cout << "Else" << endl;
		unsigned int i;
		//cout << numOfSons << endl;
		for(i=0;i<numOfSons;i++){

			//cout << "son[i] =" << sons[i] << endl;
			if (sons.at(i)!=NULL){
			//cout << "Adreese Son i = " << sons[1] << endl;
			vector<unsigned int> *tmpidx = new vector<unsigned int>;
			sons.at(i)->getIndicesOfTree(tmpidx);
			//cout << "tmpidx " << tmpidx << "   size=" << tmpidx->size() <<  endl;

			for(unsigned int k=0; k < tmpidx->size();k++){
				//cout << "Element" << tmpidx->at(k) << endl;
				index_out->push_back(tmpidx->at(k)); //concatenate
			}//for

			//index_out->insert( index_out->end(), tmpidx->begin(), tmpidx->end() );
			}
		}//for
	}//else
}//getIndices

/**
 * output: all indices which lie in the boxes l,m of the tree X
 *
 * @param index_out
 */
void tree1D::getIndicesInBox(vector <unsigned int> * index_out,unsigned int l_input, unsigned int m_input){
//Output: indices from the node matrix, which lie in X_l,m
	//cout << "getIndicesInBox" << endl;

	if (l_input > L){
		cout << "Level " << l << " exceeds range!" << endl;
		index_out=NULL;
	}

	else if(m_input > pow(2,l_input)-1){
		cout << "Parameter m= " << m_input << " exceeds range!" << endl;
		index_out=NULL;
	}

	else{
		if (l==l_input && m==m_input){
				getIndicesOfTree(index_out);
		}
		else{
			for (unsigned int k=0;k<numOfSons;k++){
				/* cout << "k" << endl;
    	   	   cout << l_input << " " << m_input << endl;
    	   	   cout << this->l << " " << this->m << endl;
    	   	   cout << " " << endl;*/
				if (sons.at(k) !=NULL){
					vector<unsigned int> *tmpidx = new vector<unsigned int>;
					sons.at(k)->getIndicesInBox(tmpidx,l_input, m_input);
					for(unsigned int k=0; k < tmpidx->size();k++){
    	   					//cout << tmpidx->at(k) << endl;
    	   					index_out->push_back(tmpidx->at(k)); //concatenate
    	   			}//for
					//index_out->insert( index_out->end(), tmpidx->begin(), tmpidx->end() );
				}//if
			}//for
		}//else
	}//else
}//IndexInBox

bool tree1D::isEmpty(void){
	if(pind !=NULL){
		return(1);
	}
	else if (numOfSons != 0){
		return(1);
	}
	else{
		return(0);
	}
}
